import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.net.CookieHandler;
import java.net.CookieManager;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

public class WebApiSampleArchives3 {
	/** URLベース */
    private static final String URL_BASE = "https://XXXXXXXX.spa-cloud.com/spa/service";
    /** 文書アーカイブ用の URL */
    private static final String ARCHIVES_URL = URL_BASE + "/archives_v3";

    /** URL ログイン */
    private static final String URL_LOGIN = URL_BASE + "/auth/login";
    /** URL ログアウト */
    private static final String URL_LOGOUT = URL_BASE + "/auth/logout";

    /** ユーザID */
    private static final String USERID = "admin";
    /** パスワード */
    private static final String PASSWORD = "admin";

    /* 改行コード */
    private static final String CRLF = "\r\n";

	public static void main(String[] args) throws Exception {

        //全てのCookie
        String cookieAll = null;
        //JSESSIONID
        String jsessionId = null;
        //XSRF-TOKEN
        String xsrfToken = null;


		//ログイン処理 START------------------------------------------------------------------------------------------

        //Cookieの処理
		CookieManager cm = new CookieManager();
		CookieHandler.setDefault(cm);

        // HttpURLConnectionの作成
        URL urlLogin = new URL(URL_LOGIN);
        HttpURLConnection conn = (HttpURLConnection) urlLogin.openConnection();

        try {
            // HttpURLConnectionの各種設定
            conn.setRequestMethod("POST");
            conn.setDoInput(true);
            conn.setDoOutput(true);
            conn.setUseCaches(false);
            conn.setAllowUserInteraction(false);
            conn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");

            // ユーザーIDとパスワードをセットする
            try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(conn.getOutputStream(), "UTF-8"))) {
                writer.write("user=" + USERID);
                writer.write("&password=" + PASSWORD);
            }

            // 正常に終了することができなかった場合
            if (conn.getResponseCode() != 200) {
                Map<String, List<String>> headers = conn.getHeaderFields();
            	StringBuffer hedaerBuf = new StringBuffer();
                for (Map.Entry<String, List<String>> entry:headers.entrySet()) {
                	String key = entry.getKey();
                	hedaerBuf.append(key + " : ");
                	List<String> list = entry.getValue();
                    for(Iterator<String>itr = list.iterator(); itr.hasNext();){
                    	String cookie = itr.next();
                    	hedaerBuf.append(cookie).append("; \r\n");
                    }
                }

                // HTTPステータスとSPA独自のエラーコードを取得できる
                int errCode = conn.getHeaderFieldInt("X-Spa-Error-Code", -1);
                System.out.println("HTTP ErrorCode[" + conn.getResponseCode() + "]");
                System.out.println("SPA ErrorCode[" + errCode + "]");
                String errMsg = conn.getHeaderField("X-Spa-Error-Message");
                if (errMsg != null) {
                    System.out.println("SPA Error Message[" + URLDecoder.decode(errMsg, "UTF-8") + "]");
                }
                throw new Exception(conn.getResponseMessage());
            }
            System.out.println("ログイン成功");

            // 正常終了時
            // セッションIDを取り出す（次のリクエスト以降で常にセット）
            Map<String, List<String>> headers = conn.getHeaderFields();
        	StringBuffer sbCookie = new StringBuffer();
            for (Map.Entry<String, List<String>> entry:headers.entrySet()) {
            	String key = entry.getKey();
            	//レスポンスヘッダからCookieを取り出す
            	if (key != null && key.equals("Set-Cookie")) {
            		List<String> list = entry.getValue();
                    for(Iterator<String>itr = list.iterator(); itr.hasNext();){
                    	String cookie = itr.next();
                    	sbCookie.append(cookie).append("; ");
                        String[] array = cookie.split(";");
                        for (String str : array) {
                        	//JSESSIONIDの値を取り出す
                        	if (str.matches("JSESSIONID=.*")) {
                            	String[] kvp = str.split("=");
                            	jsessionId = kvp[1];
                        	}
                        	//XSRF-TOKENの値を取り出す
                            if (str.matches("XSRF-TOKEN=.*")) {
                            	String[] kvp = str.split("=");
                            	xsrfToken = kvp[1];
                            }
                        }

                    }
                }
            }
            cookieAll = sbCookie.toString();
            System.out.println("cookieAll = " + cookieAll);
            System.out.println("JSESSIONID = " + jsessionId);
            System.out.println("XSRF-TOKEN = " + xsrfToken);

            //ログイン処理 END------------------------------------------------------------------------------------------

            //アーカイブ処理 START--------------------------------------------------------------------------------------
            System.out.println("アーカイブ処理開始");
            String path = "/sample";
            File file = new File("c:\\sample.pdf");
            String name = file.getName();

            // HttpURLConnectionの作成
            URL urlArchive = new URL(ARCHIVES_URL);
            conn = (HttpURLConnection) urlArchive.openConnection();

            // HttpURLConnectionの各種設定
            conn.setRequestMethod("POST");
            conn.setDoInput(true);
            conn.setDoOutput(true);
            conn.setUseCaches(false);
            conn.setAllowUserInteraction(false);

            // XSRF-TOKENとX-Requested-Withをセット(リクエストのヘッダに)
            conn.setRequestProperty("x-xsrf-token", xsrfToken);
            conn.setRequestProperty("x-requested-with", "XMLHttpRequest");

            // マルチパートメッセージの作成
            String boundary = "----SpaFormBoundary" + Long.toString(System.currentTimeMillis());
            conn.setRequestProperty("Content-Type", "multipart/form-data; boundary=" + boundary);

            try (DataOutputStream dos = new DataOutputStream(new BufferedOutputStream(conn.getOutputStream()))) {
            	//path 保存先パス指定
                postParam(dos, boundary, "path", path);
                //name ファイル名指定
                postParam(dos, boundary, "name", name);
                //カスタムプロパティセット
                postcustomProperties(dos, boundary, 1000, "20220114", "ウイングアーク1st株式会社");
                //file ファイルデータそのもの送信
                postFile(dos, boundary, "file", file.getName(), file);
                dos.writeBytes("--" + boundary + "--" + CRLF);
            }
            System.out.println("送信完了");

            // 正常に終了することができなかった場合
            if (conn.getResponseCode() != 200) {
                headers = conn.getHeaderFields();
            	StringBuffer hedaerBuf = new StringBuffer();
                for (Map.Entry<String, List<String>> entry:headers.entrySet()) {
                	String key = entry.getKey();
                	hedaerBuf.append(key + " : ");
                	List<String> list = entry.getValue();
                    for(Iterator<String>itr = list.iterator(); itr.hasNext();){
                    	String cookie = itr.next();
                    	hedaerBuf.append(cookie).append("; \r\n");
                    }
                }

                // HTTPステータスとSPA独自のエラーコードを取得できる
                int errCode = conn.getHeaderFieldInt("X-Spa-Error-Code", -1);
                System.out.println("HTTP ErrorCode[" + conn.getResponseCode() + "]");
                System.out.println("SPA ErrorCode[" + errCode + "]");
                String errMsg = conn.getHeaderField("X-Spa-Error-Message");
                if (errMsg != null) {
                    System.out.println("SPA Error Message[" + URLDecoder.decode(errMsg, "UTF-8") + "]");
                }
                throw new Exception(conn.getResponseMessage());
            }
            System.out.println("アーカイブ成功");
            //アーカイブ処理 END ---------------------------------------------------------------------------------------

            //アーカイブ後のレスポンス処理 JSONの値取得 START-----------------------------------------------------------
            InputStream in = conn.getInputStream();

            // 正常終了した場合はJSONを取得できる
            ObjectMapper mapper = new ObjectMapper();
            JsonNode document = mapper.readTree(conn.getInputStream());
            //id
            String id = document.get("id").asText();
            //direct_url
            String direct_url = document.get("direct_url").asText();

            System.out.println("id: " + id);
            System.out.println("direct_url: " + direct_url);
            in.close();


            //ログアウト処理 START--------------------------------------------------------------------------------------

            // HttpURLConnectionの作成
            URL urlLogout = new URL(URL_LOGOUT);
            conn = (HttpURLConnection) urlLogout.openConnection();

            try {
                // HttpURLConnectionの各種設定
                conn.setRequestMethod("POST");
                conn.setDoInput(true);
                conn.setDoOutput(true);
                conn.setUseCaches(false);
                conn.setAllowUserInteraction(false);

                // XSRF-TOKENとX-Requested-Withをセット(リクエストのヘッダに)
                conn.setRequestProperty("x-xsrf-token", xsrfToken);
                conn.setRequestProperty("x-requested-with", "XMLHttpRequest");

                // 正常に終了することができなかった場合
                if (conn.getResponseCode() != 200) {
                    // HTTPステータスとSPA独自のエラーコードを取得できる
                    int errCode = conn.getHeaderFieldInt("X-Spa-Error-Code", -1);
                    System.out.println("HTTP ErrorCode[" + conn.getResponseCode() + "]");
                    System.out.println("SPA ErrorCode[" + errCode + "]");
                    String errMsg = conn.getHeaderField("X-Spa-Error-Message");
                    if (errMsg != null) {
                        System.out.println("SPA Error Message[" + URLDecoder.decode(errMsg, "UTF-8") + "]");
                    }
                    throw new Exception(conn.getResponseMessage());
                }
                System.out.println("logout succeed.");
            } finally {
                // コネクションの切断
                conn.disconnect();
            }

            //ログアウト処理 END----------------------------------------------------------------------------------------

        } finally {
            // コネクションの切断
            conn.disconnect();
        }

	}

    /**
     * カスタムプロパティ書き込み処理　（取引金額、取引日付、取引先）
     *
     * @param dos DataOutputStream
     * @param boundary boundary文字列
     * @param amount 取引金額
     * @param date 取引日付 yyyyMMdd
     * @param customer 取引先
     * @throws IOException 書き込み処理に失敗した場合
     */
    private static void postcustomProperties(DataOutputStream dos, String boundary, int amount , String date, String customer)
            throws IOException {

        dos.writeBytes("--" + boundary + CRLF);
        dos.writeBytes("Content-Disposition: form-data; name=\"" + "customProperties" + "\"" + CRLF);
        dos.writeBytes("Content-Type: application/json; charset=UTF-8" + CRLF);
        dos.writeBytes(CRLF);
        dos.writeBytes("{");
        dos.writeBytes("  \"customProperties\": {");

        //SPA_amount 取引金額
        dos.writeBytes("    \"50\": \"" + amount + "\",");
        //SPA_date 取引日付
        dos.writeBytes("    \"51\": \"" + date + "\",");
        //SPA_customer 取引先
        dos.writeBytes("    \"52\": \"");
        dos.write(customer.getBytes(StandardCharsets.UTF_8));
        dos.writeBytes("\"");

        dos.writeBytes("  }");
        dos.writeBytes("}");
        dos.writeBytes(CRLF);
    }

    /**
     * フォームデータの書き込み処理
     *
     * @param dos DataOutputStream
     * @param boundary boundary文字列
     * @param name フィールド名
     * @param value 値
     * @throws IOException 書き込み処理に失敗した場合
     */
    private static void postParam(DataOutputStream dos, String boundary, String name, String value)
            throws IOException {
        dos.writeBytes("--" + boundary + CRLF);

        dos.writeBytes("Content-Disposition: form-data; name=\"" + name + "\"" + CRLF);
        dos.writeBytes(CRLF);
        dos.write(value.getBytes(StandardCharsets.UTF_8));
        dos.writeBytes(CRLF);
    }

    /**
     * フォームデータのファイルへの書き込み処理
     *
     * @param dos DataoutputStream
     * @param boundary boundary文字列
     * @param name フィールド名
     * @param value 値
     * @param file 送信するファイル
     * @throws IOException 書き込み処理に失敗した場合
     */
    private static void postFile(DataOutputStream dos, String boundary, String name, String value, File file)
            throws IOException {

        dos.writeBytes("--" + boundary + CRLF);

        StringBuilder sb = new StringBuilder();
        sb.append("Content-Disposition: form-data; ");
        sb.append("name=\"").append(name).append("\"; ");
        sb.append("filename=\"").append(value).append("\"");
        sb.append(CRLF);
        dos.write(sb.toString().getBytes(StandardCharsets.UTF_8));

        dos.writeBytes("Content-Type: application/pdf" + CRLF);
        dos.writeBytes(CRLF);

        // ファイルの書き込み
        try (BufferedInputStream fis = new BufferedInputStream(new FileInputStream(file))) {
            byte[] buffer = new byte[4096];
            int len = -1;
            while ((len = fis.read(buffer)) != -1) {
                dos.write(buffer, 0, len);
            }
        }
        dos.writeBytes(CRLF);
    }
}
